/*
 * $Id: wcfxs.c 67 2005-01-19 19:32:01Z riz $
 *
 *
 * NetBSD driver for Wildcard X100P FXO board from Digium, Inc.
 *
 * Original code written by:
 *	Mark Spencer <markster@linux-support.net>
 *	Matthew Fredrickson <creslin@linux-support.net>
 *
 * Copyright (C) 2001, Linux Support Services, Inc.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * FreeBSD port by
 *	Alexander Timoshenko <gonzo@portaone.com>, PortaOne software
 *
 * NetBSD port, and code cleanup and rewriting 
 * 	Copyright (c) 2005, Red Crow Group LLC
 *	written by Jeff Rizzo <riz@redcrowgroup.com>
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$Id: wcfxs.c 67 2005-01-19 19:32:01Z riz $");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/proc.h>
#include <sys/uio.h>
#include <sys/malloc.h>
#include <sys/device.h>
#include <sys/conf.h>

#include <uvm/uvm_extern.h>	/* for PAGE_SIZE */

#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#include <dev/pci/pcidevs.h>

#ifdef _LKM
#include <zaptel.h>
#include "wcfxoreg.h"
#include "wcfxovar.h"
#else
#include <dev/zaptel/zaptel/zaptel.h>
#include <dev/zaptel/wcfxo-netbsd/wcfxoreg.h>
#include <dev/zaptel/wcfxo-netbsd/wcfxovar.h>
#endif


#define DUMP_REG(name)	do \
{ int x; \
	printf("%s: reg dump ============\n", name); \
	for (x=0; x<sizeof(sc->sc_readregs) ;x++) {\
		printf("%s: Reg %d: %02x\n", name, x, sc->sc_readregs[x]); \
	} \
} while (0)

int wcfxo_match(struct device *, struct cfdata *, void *);
static void wcfxo_attach(struct device *, struct device *, void *);
static int wcfxo_detach(struct device *, int);
static void wcfxo_finish_config(struct device *);

struct device *zaptel_attach_mi(struct device *);
int wcfxo_open(struct zt_chan *chan);
static int wcfxo_close(struct zt_chan *chan);
static int wcfxo_watchdog(struct zt_span *span, int event);

static int wcfxo_intr(void *);
static int wcfxo_initialize(struct wcfxo_softc *);

static void wcfxo_start_dma(struct wcfxo_softc *sc);
static void wcfxo_restart_dma(struct wcfxo_softc *sc);
static void wcfxo_stop_dma(struct wcfxo_softc *sc);
static void wcfxo_reset_tdm(struct wcfxo_softc *sc);
static void wcfxo_release(struct wcfxo_softc *sc);
void wcfxo_generate_description(struct wcfxo_softc * sc);
static int wcfxo_hooksig(struct zt_chan *chan, zt_txsig_t txsig);
static int wcfxo_setreg(struct wcfxo_softc *, unsigned char, unsigned char);

int wcfxodebug = 1;
#ifdef DEBUG
#define DPRINTF(m, x)   if (wcfxodebug >= (m)) printf x
#define DEBUG_DO(m, x)	if (wcfxodebug >= (m)) do { x ; } while (0)
#else
#define DPRINTF(m, x)   /* nop */
#define DEBUG_DO(m, x)	/* nop */
#endif
static int monitor = 0;
static int quiet = 0;
static int boost = 0;
static int opermode = 2;
static int batterycheck = 1;
static char * card_configuration = NULL;

CFATTACH_DECL(wcfxo, sizeof(struct wcfxo_softc), 
	wcfxo_match, wcfxo_attach, wcfxo_detach, NULL);

/* descriptions for the boards which might match */
const static struct wcfxo_desc wcx100p = { "Wildcard X100P",
	FLAG_INVERTSER | FLAG_USE_XTAL | FLAG_DOUBLE_CLOCK };
const static struct wcfxo_desc wcx101p = { "Wildcard X101P",
	FLAG_USE_XTAL | FLAG_DOUBLE_CLOCK };
const static struct wcfxo_desc generic = { "Generic X100P Clone",
	FLAG_USE_XTAL | FLAG_DOUBLE_CLOCK };

/* board identification table */
static struct wcfxo_ident wcfxo_ident_table[] = {
	/* vendor, device, subvendor, card info struct */
	{ 0xe159, 0x0001, 0x8085, &wcx101p },
	{ 0xe159, 0x0001, 0x8086, &generic },
	{ 0x1057, 0x5608, -1, &wcx100p },
	{ 0, 0, 0, NULL }
};

static int we_care_regs[] = { 5, 6, 9, 11, 12, 13, 17, 19, };

int
wcfxo_match(struct device *parent, struct cfdata *match, void *aux)
{
	struct pci_attach_args *pa = aux;
	pcireg_t subdev;
	struct wcfxo_ident *id;

	DPRINTF(1,("%s\n",__func__));
	/* read the subdevice information */
	subdev = pci_conf_read(pa->pa_pc, pa->pa_tag, PCI_SUBSYS_ID_REG);

	for (id = wcfxo_ident_table; id->device_info != NULL; id++) {
		if (id->vendor == PCI_VENDOR(pa->pa_id) &&
			id->product == PCI_PRODUCT(pa->pa_id) &&
			(id->subvendor == PCI_VENDOR(subdev) ||
			id->subvendor == (pci_vendor_id_t) -1)) {

			printf("FXO device: ");
			printf("vendor=%x ",PCI_VENDOR(pa->pa_id));
			printf("product=%x ",PCI_PRODUCT(pa->pa_id));
			printf("subvendor=%x\n",PCI_VENDOR(subdev));
			return (1);
		}
	}
	return (0);
}

/* wcfxo_dma_malloc() returns 0 on failure, 1 on success */
static int
wcfxo_dma_malloc(struct wcfxo_softc *sc, struct wcfxo_dma *dma)
{
	int error;

	DPRINTF(1,("%s\n",__func__));
	/* set up read buffer first */
	if ((error = bus_dmamem_alloc(sc->sc_dmat, WC_SIZE, PAGE_SIZE, 0,
	    &dma->dma_seg, 1, &dma->dma_nseg, BUS_DMA_NOWAIT)) != 0) {
		aprint_error(
		    "%s: unable to allocate readchunk, error = %d\n",
		    sc->sc_dv.dv_xname, error);
		goto fail_0;
	}

	if ((error = bus_dmamem_map(sc->sc_dmat, &dma->dma_seg,
	    dma->dma_nseg, WC_SIZE, (caddr_t *)&dma->chunk,
	    BUS_DMA_NOWAIT)) != 0) {
		aprint_error(
		    "%s: unable to map read chunk, error = %d\n",
		    sc->sc_dv.dv_xname, error);
		goto fail_1;
	}

	if ((error = bus_dmamap_create(sc->sc_dmat, WC_SIZE, 1, WC_SIZE,
								   0, 0, &dma->dma_map)) != 0) {
		aprint_error("%s: unable to create read DMA map, "
		    "error = %d\n",sc->sc_dv.dv_xname, error);
		goto fail_2;
	}

	if ((error = bus_dmamap_load(sc->sc_dmat, dma->dma_map,
	     dma->chunk, WC_SIZE, NULL, BUS_DMA_NOWAIT)) != 0) {
		aprint_error(
		    "%s: unable to load read DMA map, error = %d\n",
		    sc->sc_dv.dv_xname, error);
		goto fail_3;
	}

	/* success! */
	return 1;

 fail_3:
	bus_dmamap_destroy(sc->sc_dmat, dma->dma_map);
 fail_2:
	bus_dmamem_unmap(sc->sc_dmat, (caddr_t)dma->chunk, WC_SIZE);
 fail_1:
	bus_dmamem_free(sc->sc_dmat, &dma->dma_seg,
	    dma->dma_nseg);
 fail_0:
	return 0;
}

static void
wcfxo_dma_free(struct wcfxo_softc *sc, struct wcfxo_dma *dma)
{
	int size = ZT_MAX_CHUNKSIZE * 2 * 2 * 2 * 4; /* XXX magic */
	DPRINTF(1,("%s\n",__func__));

	bus_dmamap_destroy(sc->sc_dmat, dma->dma_map);
	bus_dmamem_unmap(sc->sc_dmat, (caddr_t)dma->chunk, size);
	bus_dmamem_free(sc->sc_dmat, &dma->dma_seg,
	    dma->dma_nseg);
}

/*
 * wcfxo_initialize() - set up various bits of the softc
 */
static int
wcfxo_initialize(struct wcfxo_softc *sc)
{
	DPRINTF(1,("%s\n",__func__));

	/* set up the Zaptel bits */
	sprintf(sc->sc_span.name, "WCFXO/%d", sc->sc_pos);
	sprintf(sc->sc_span.desc, "%s Board %d", sc->sc_variety,
	    sc->sc_pos + 1);
	sprintf(sc->sc_chan.name, "WCFXO/%d/%d",sc->sc_pos, 0);
	sc->sc_chan.sigcap = ZT_SIG_FXSKS | ZT_SIG_FXSLS | ZT_SIG_SF;
	sc->sc_chan.chanpos = 1;
	sc->sc_span.chans = &sc->sc_chan;
	sc->sc_span.channels = 1;
	sc->sc_span.hooksig = wcfxo_hooksig;
	sc->sc_span.open = wcfxo_open;
	sc->sc_span.close = wcfxo_close;
	sc->sc_span.flags = ZT_FLAG_RBS;
	sc->sc_span.deflaw = ZT_LAW_MULAW;
	sc->sc_span.watchdog = wcfxo_watchdog;
	/* TODO: waitq init_waitqueue_head(&sc->sc_span.maintq); */

	sc->sc_span.pvt = sc;
	sc->sc_chan.pvt = sc;

	if (zt_register(&sc->sc_span, 0)) {
	    printf("Unable to register span with zaptel\n");
	    return -1;
	}
	return 0;
}

static int wcfxo_init_daa(struct wcfxo_softc *sc)
{
	unsigned char reg15;
	/* This must not be called in an interrupt */
	/* We let things settle for a bit */
	tsleep(sc, PZERO | PCATCH, "waitintr", 10);

	DUMP_REG(__func__);
	/* Soft-reset it */
	wcfxo_setreg(sc, 0x1, 0x80);

	/* Let the reset go */
	/* set_current_state(TASK_UNINTERRUPTIBLE);
	schedule_timeout(1 + (ZT_CHUNKSIZE * HZ) / 800); */
	/* printf("+delay 2(%d)\n", (1 + (ZT_CHUNKSIZE * HZ) / 800)*10);
	DELAY((1 + (ZT_CHUNKSIZE * HZ) / 800)*10);
	printf("-delay 2\n"); */
	tsleep(sc, PZERO, "waitintr", (1 + (ZT_CHUNKSIZE*hz)/800)*10 );

	/* We have a clock at 18.432 Mhz, so N1=1, M1=2, CGM=0 */
	wcfxo_setreg(sc, 0x7, 0x0); /* This value is N1 - 1 */
	wcfxo_setreg(sc, 0x8, 0x1); /* This value is M1 - 1 */
	/* We want to sample at 8khz, so N2 = 9, M2 = 10 (N2-1, M2-1) */
	wcfxo_setreg(sc, 0x9, 0x89);

	/* Wait until the PLL's are locked. Time is between 100 uSec and 1 mSec */
	/*set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(1 + HZ/1000 + (ZT_CHUNKSIZE * HZ) / 800); */
	/* printf("delay 3\n");
	DELAY((1 + HZ/1000 + (ZT_CHUNKSIZE * HZ) / 800)*10);  */
	tsleep(sc, PZERO | PCATCH, "waitintr", 
		(1 + hz/1000 + (ZT_CHUNKSIZE * hz) / 800)*10);

	/* No additional ration is applied to the PLL and faster lock times
	 * are possible */
	wcfxo_setreg(sc, 0xa, 0x0);
	/* Enable off hook pin */
	wcfxo_setreg(sc, 0x5, 0x0a);
	if (monitor) {
		/* Enable ISOcap and external speaker and charge pump if present */
		wcfxo_setreg(sc, 0x6, 0x80);
	} else {
		/* Enable ISOcap and charge pump if present (leave speaker disabled) */
		wcfxo_setreg(sc, 0x6, 0xe0);
	}

	/* Wait a couple of jiffies for our writes to finish */
	/*set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(1 + (ZT_CHUNKSIZE * HZ) / 800);*/
	/* printf("delay 4\n");
	 DELAY((1 + (ZT_CHUNKSIZE * HZ) / 800)*10); */
	tsleep(sc, PZERO | PCATCH, "waitintr", (1 + (ZT_CHUNKSIZE * hz) / 800)*10);

	reg15 = 0;
	/* Go ahead and attenuate transmit signal by 6 db */
	if (quiet) {
		printf("wcfxo: Attenuating transmit signal for quiet operation\n");
		reg15 |= (quiet & 0x3) << 4;
	}

	if (boost) {
		printf("wcfxo: Boosting receive signal\n");
		reg15 |= (boost & 0x3);
	}

	wcfxo_setreg(sc, 0xf, 0x10);
	DUMP_REG(__func__);

	/* Didn't get it right.  Register 9 is still garbage */
	if (sc->sc_readregs[0x9] != 0x89)
		return -1;
	return 0;
}

static int wcfxo_set_daa_mode(struct wcfxo_softc *sc)
{
	/* Set country specific parameters (OHS, ACT, DCT, RZ, RT, LIM, VOL) */
	int reg16 = ((fxo_modes[opermode].ohs & 0x1) << 6) |
				((fxo_modes[opermode].act & 0x1) << 5) |
				((fxo_modes[opermode].dct & 0x3) << 2) |
				((fxo_modes[opermode].rz & 0x1) << 1) |
				((fxo_modes[opermode].rt & 0x1) << 0);
	int reg17 = ((fxo_modes[opermode].lim & 0x3) << 3);
	int reg18 = ((fxo_modes[opermode].vol & 0x3) << 3);

	wcfxo_setreg(sc, 16, reg16);
	wcfxo_setreg(sc, 17, reg17);
	wcfxo_setreg(sc, 18, reg18);


	/* Wait a couple of jiffies for our writes to finish */
	tsleep(sc, PZERO, "writefinish", 1 + (ZT_CHUNKSIZE * hz) / 800);

	printf("wcfxo: DAA mode is '%s'\n", fxo_modes[opermode].name);
	return 0;
}


static int
wcfxo_hardware_init(struct wcfxo_softc *sc)
{
	unsigned char c;
	u_int32_t write_dma_addr,read_dma_addr;

	/* I'd love to get rid of some of these magic numbers - riz */
	DPRINTF(1,("%s\n",__func__));

	/* Reset PCI Interface chip and registers */
	WC_WRITE1(WC_CNTL, 0x0e);
	if (sc->sc_flags & FLAG_RESET_ON_AUX5) {
		/* Set hook state to on hook for when we switch.
		   Make sure reset is high. */
		WC_WRITE1(WC_AUXD, 0x34);
	} else {
		/* Set hook state to on hook for when we switch. */
		WC_WRITE1(WC_AUXD, 0x24);
	}

	/* Set all to outputs except AUX 4, which is an input */
	DPRINTF(1,("%s: writing 0x%x\n",__func__, (~AUX4_IN & 0xff)));
	WC_WRITE1(WC_AUXC, (~AUX4_IN & 0xff));

	/* Wait 1/4 sec */
	//tsleep(sc, PZERO, "wcfxo_hardware_init", (hz / 4) + 1);

	/* Back to normal, with automatic DMA wraparound */
	WC_WRITE1(WC_CNTL, 0x01);

	/* Make sure serial port and DMA are out of reset */
	c = WC_READ1(WC_CNTL);
	c &= 0xf9;
	WC_WRITE1(WC_CNTL, c);

	/* Configure serial port for MSB->LSB operation */
	if (sc->sc_flags & FLAG_DOUBLE_CLOCK)
		WC_WRITE1(WC_SERCTL, (SERIN_MSB_FIRST | SEROUT_MSB_FIRST
			| SER2XDCLK));
	else
		WC_WRITE1(WC_SERCTL, (SERIN_MSB_FIRST | SEROUT_MSB_FIRST));

	if (sc->sc_flags & FLAG_USE_XTAL) {
		/* Use the crystal oscillator */
		WC_WRITE1(WC_AUXFUNC, WC_CKO_ENABLE);
	}

	/* Delay FSC by 2 so it's properly aligned */
	WC_WRITE1(WC_FSCDELAY, 0x2);

	/* Setup DMA Addresses - XXX riz */
	write_dma_addr = vtophys((vaddr_t)sc->sc_write.chunk);
	WC_WRITE4(WC_DMAWS, (unsigned int)write_dma_addr); /* write start */
	WC_WRITE4(WC_DMAWI, (unsigned int)write_dma_addr +
	    ZT_CHUNKSIZE * 8 - 4);	/* Middle (interrupt) */
	WC_WRITE4(WC_DMAWE, (unsigned int)write_dma_addr +
	    ZT_CHUNKSIZE * 16 - 4);	/* End */

	read_dma_addr = vtophys((vaddr_t)sc->sc_read.chunk);
	WC_WRITE4(WC_DMARS, (unsigned int)read_dma_addr); /* write start */
	WC_WRITE4(WC_DMARI, (unsigned int)read_dma_addr +
	    ZT_CHUNKSIZE * 8 - 4);	/* Middle (interrupt) */
	WC_WRITE4(WC_DMARE, (unsigned int)read_dma_addr +
	    ZT_CHUNKSIZE * 16 - 4);	/* End */

	/* Clear interrupts */
	WC_WRITE1(WC_INTSTAT, 0xff);

	/* wait another 1/2 sec */
//	tsleep(sc, PZERO, "wcfxo_hardware_init", (hz / 2) + 1);

	return 0;
}

static void wcfxo_enable_interrupts(struct wcfxo_softc *sc)
{
	/* Enable interrupts (we care about all of them) */
	WC_WRITE1(WC_MASK0, 0x3f);
	/* No external interrupts */
	WC_WRITE1(WC_MASK1, 0x00);
	DPRINTF(1,("%s\n",__func__));
}

static void wcfxo_restart_dma(struct wcfxo_softc *sc)
{
	/* Reset Master and TDM */
	WC_WRITE1(WC_CNTL, 0x01);
	WC_WRITE1(WC_OPER, 0x01);
	DPRINTF(1,("%s\n",__func__));
}

static void wcfxo_start_dma(struct wcfxo_softc *sc)
{
	/* Reset Master and TDM */
	WC_WRITE1(WC_CNTL, 0x0f);
	tsleep(sc, PZERO | PCATCH, "startdma", 1);
	WC_WRITE1(WC_CNTL, 0x01);
	WC_WRITE1(WC_OPER, 0x01);
	DPRINTF(1,("%s\n",__func__));
}

static void wcfxo_stop_dma(struct wcfxo_softc *sc)
{
	WC_WRITE1(WC_OPER, 0x00);
	DPRINTF(1,("%s\n",__func__));
}

static void wcfxo_reset_tdm(struct wcfxo_softc *sc)
{
	/* Reset TDM */
	WC_WRITE1(WC_CNTL, 0x0f);
}

static void wcfxo_disable_interrupts(struct wcfxo_softc *sc)	
{
	WC_WRITE1(WC_MASK0, 0x00);
	WC_WRITE1(WC_MASK1, 0x00);
	DPRINTF(1,("%s\n",__func__));
}

static __inline void wcfxo_transmitprep(struct wcfxo_softc *sc,
		unsigned char ints)
{
	volatile unsigned int *writechunk;
	int x;
	int written = 0;
	unsigned short cmd;

	/* if nothing to transmit, have to do the zt_transmit() anyway */
	if (!(ints & 3)) {
		/* Calculate Transmission */
		zt_transmit(&sc->sc_span);
		return;
	}

	/* Remember what it was we just sent */
	memcpy(sc->sc_lasttx, sc->sc_chan.writechunk, ZT_CHUNKSIZE);

	if (ints & 0x01) 
		/* Write is at interrupt address.  Start writing from
		   normal offset */
		writechunk = sc->sc_write.chunk;
	else 
		writechunk = sc->sc_write.chunk + ZT_CHUNKSIZE * 2;

	/* Calculate Transmission */
	zt_transmit(&sc->sc_span);

	bus_dmamap_sync(sc->sc_dmat, sc->sc_write.dma_map, 0,
		sc->sc_write.dma_map->dm_mapsize, BUS_DMASYNC_PREREAD);
	for (x = 0; x < ZT_CHUNKSIZE; x++) {
	    /* Send a sample as a 32-bit word, and be sure
	       to indicate that a command follows */
	    if (sc->sc_flags & FLAG_INVERTSER)
		writechunk[x << 1] = htole32( ~((unsigned short)(ZT_XLAW(
			sc->sc_chan.writechunk[x], (&sc->sc_chan))) | 0x1)
				<< 16);
	    else
		writechunk[x << 1] = htole32( ((unsigned short)(ZT_XLAW(
			sc->sc_chan.writechunk[x], (&sc->sc_chan))) | 0x1)
				<< 16);

	    /* We always have a command to follow our signal */
	    if (!sc->sc_regs[x].flags) {
		/* Fill in an empty register command with a read for a
		   potentially useful register */
		sc->sc_regs[x].flags = FLAG_READ;
		sc->sc_regs[x].reg = we_care_regs[sc->sc_readpos];
		sc->sc_regs[x].index = sc->sc_readpos;
		sc->sc_readpos++;
		if (sc->sc_readpos >= (sizeof we_care_regs /
			sizeof we_care_regs[0])) {
		   sc->sc_allread = 1;
		   sc->sc_readpos = 0;
		}
	    }
	    /* Prepare the command to follow it */
	    switch(sc->sc_regs[x].flags) {
	    case FLAG_READ:
		cmd = (sc->sc_regs[x].reg | 0x20) << 8;
		break;
	    case FLAG_WRITE:
		cmd = (sc->sc_regs[x].reg  << 8) | 
			(sc->sc_regs[x].value & 0xff);
		written = 1;
		/* Wait at least 4 samples before reading */
		sc->sc_ignoreread = 4;
		break;
	    default:
		printf("%s: Huh? No read or write??\n",sc->sc_dv.dv_xname);
		cmd = 0;
		break;
	    }
	    /* Set up the write chunk */
	    if (sc->sc_flags & FLAG_INVERTSER)
		writechunk[(x << 1) + 1] = htole32(~(cmd << 16));
	    else
		writechunk[(x << 1) + 1] = htole32(cmd << 16);
	}

	if (written)
		sc->sc_readpos = 0;
	sc->sc_wregcount = 0;

	for (x=0; x < ZT_CHUNKSIZE; x++) {
	    /* Rotate through registers */
	    sc->sc_oldregs[x] = sc->sc_regs[x];
	    sc->sc_regs[x].flags = FLAG_EMPTY;
	}

	bus_dmamap_sync(sc->sc_dmat, sc->sc_write.dma_map, 0,
		sc->sc_write.dma_map->dm_mapsize, BUS_DMASYNC_POSTREAD);
}


void wcfxo_generate_description(struct wcfxo_softc * sc)
{
	DPRINTF(1,("%s\n",__func__));
	/* prepare description */

	/* 80(name) + 40(description) + 25 per port description 
	 * + 20 just in case
	 */
	if(!card_configuration)
		card_configuration = malloc(80 + 40 + 25 + 20, 
						M_DEVBUF, M_NOWAIT);
	snprintf(card_configuration, 120 + 25, "%s\n%s\n", 
					sc->sc_span.name, sc->sc_span.desc);
	snprintf(card_configuration, 120 + 25, 
		"%sPort #%d -- FXO [%s]\n", card_configuration, 1,
		sc->sc_span.alarms ? "off" : "on");
}


static __inline void wcfxo_receiveprep(struct wcfxo_softc *sc, unsigned char ints)
{
	volatile unsigned int *readchunk;
	int x;
	int realreg;
	int realval;
	int sample;

	if (ints & 0x04)
		/* Read is at interrupt address.
		   Valid data is available at normal offset */
		readchunk = sc->sc_read.chunk;
	else
		readchunk = sc->sc_read.chunk + ZT_CHUNKSIZE * 2;

	bus_dmamap_sync(sc->sc_dmat, sc->sc_read.dma_map, 0,
		sc->sc_read.dma_map->dm_mapsize, BUS_DMASYNC_PREWRITE);

	/* Keep track of how quickly our peg alternates */
	sc->sc_pegtimer += ZT_CHUNKSIZE;
	sc->sc_alt = 0;

	for (x=0; x < ZT_CHUNKSIZE; x++) {
	    if (sc->sc_oldregs[x].flags == FLAG_READ && !sc->sc_ignoreread) {
		realreg = we_care_regs[(sc->sc_regs[x].index +
			sc->sc_regoffset) % (sizeof we_care_regs /
			sizeof we_care_regs[0])];
		realval = readchunk[(x << 1) + sc->sc_alt];
		realval = (le32toh(realval) >> 16) & 0xff;

		if ((realval == 0x89) && (realreg != 0x9)) {
		    /* XXX WTF are these numbers? - riz */
		    /* Some sort of slippage, correct for it */
		    while(realreg != 0x9) {
			realreg = we_care_regs[(sc->sc_regs[x].index +
			    ++sc->sc_regoffset) % (sizeof we_care_regs /
			    sizeof we_care_regs[0])];
			sc->sc_regoffset = sc->sc_regoffset %
			    (sizeof we_care_regs / sizeof we_care_regs[0]);
		    }
		    DPRINTF(1,("%s: New regoffset: %d\n", sc->sc_dv.dv_xname,
			sc->sc_regoffset));
		}
		/* Receive into the proper register */
		sc->sc_readregs[realreg] = realval;
	    }
	    /* Look for pegging to indicate ringing */
	    sample = (short)(le32toh(readchunk[(x << 1) + (1 - sc->sc_alt)])
		>> 16);
	    if ((sample > 32000) && (sc->sc_peg != 1)) {
		if ((sc->sc_pegtimer < PEGTIME) &&
			(sc->sc_pegtimer > MINPEGTIME))
		    sc->sc_pegcount++;
		sc->sc_pegtimer = 0;
		sc->sc_peg = 1;
	    } else if ((sample < -32000) && (sc->sc_peg != -1)) {
		if ((sc->sc_pegtimer < PEGTIME) &&
			(sc->sc_pegtimer > MINPEGTIME))
		    sc->sc_pegcount++;
		sc->sc_pegtimer = 0;
		sc->sc_peg = -1;
	    }
	    sc->sc_chan.readchunk[x] = ZT_LIN2X((sample), (&sc->sc_chan));
	}
	if (sc->sc_pegtimer > PEGTIME) {
	    /* Reset pegcount if our timer expires */
	    sc->sc_pegcount = 0;
	}
	/* Decrement debouncer if appropriate */
	if (sc->sc_ringdebounce)
	    sc->sc_ringdebounce--;
	if (!sc->sc_offhook && !sc->sc_ringdebounce) {
	    if (!sc->sc_ring && (sc->sc_pegcount > PEGCOUNT)) {
		/* It's ringing */
		DPRINTF(1,("%s: RING!\n", sc->sc_dv.dv_xname));
		zt_hooksig(&sc->sc_chan, ZT_RXSIG_RING);
		sc->sc_ring = 1;
	    }
	    if (sc->sc_ring && !sc->sc_pegcount) {
		/* No more ring */
		DPRINTF(1,("%s: NO RING!\n", sc->sc_dv.dv_xname));
		zt_hooksig(&sc->sc_chan, ZT_RXSIG_OFFHOOK);
		sc->sc_ring = 0;
	    }
	}
	if (sc->sc_ignoreread)
	    sc->sc_ignoreread--;

	/* Do the echo cancellation.  We are echo cancelling against
	   what we sent two chunks ago. */
	zt_ec_chunk(&sc->sc_chan, sc->sc_chan.readchunk, sc->sc_lasttx);

	/* Receive the result */
	zt_receive(&sc->sc_span);

	bus_dmamap_sync(sc->sc_dmat, sc->sc_read.dma_map, 0,
		sc->sc_read.dma_map->dm_mapsize, BUS_DMASYNC_POSTWRITE);
}

static int
wcfxo_setreg(struct wcfxo_softc *sc, unsigned char reg, unsigned char value)
{
	int x;
	if (sc->sc_wregcount < ZT_CHUNKSIZE) {
	    x = sc->sc_wregcount;
	    sc->sc_regs[x].reg = reg;
	    sc->sc_regs[x].value = value;
	    sc->sc_regs[x].flags = FLAG_WRITE;
	    sc->sc_wregcount++;
	    return 0;
	}
	printf("%s: Out of space to write register %02x with %02x\n",
		sc->sc_dv.dv_xname, reg, value);
	return -1;
}

static void
wcfxo_attach(struct device *parent, struct device *self, void *aux)
{
	struct wcfxo_softc *sc = (struct wcfxo_softc *) self;
	struct pci_attach_args *pa = aux;
	pci_chipset_tag_t pc = pa->pa_pc;
	pcireg_t subdev;
	pci_intr_handle_t ih;
	const char *intrstr;
	struct wcfxo_ident *id;
	int error;


	/* read the subdevice information - XXX code duplication */
	subdev = pci_conf_read(pc, pa->pa_tag, PCI_SUBSYS_ID_REG);

	for (id = wcfxo_ident_table ; id->device_info != NULL; id++)
		if (id->vendor == PCI_VENDOR(pa->pa_id) &&
			id->product == PCI_PRODUCT(pa->pa_id) &&
			(id->subvendor == PCI_VENDOR(subdev) ||
			id->subvendor == (pci_vendor_id_t) -1))
			break;

	if (id->device_info == NULL) {
		printf("Card info not found\n");
		return;
	}

	printf(": %s\n",id->device_info->name);

	/*
	 * Make sure bus mastering, ioports, and memory mapping are enabled.
	 */
	pci_conf_write(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG,
		pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG) |
		PCI_COMMAND_MASTER_ENABLE | PCI_COMMAND_IO_ENABLE |
		PCI_COMMAND_MEM_ENABLE);

	if (!(pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG) &
		PCI_COMMAND_IO_ENABLE)) {
		printf("%s: failed to enable I/O ports\n",
			sc->sc_dv.dv_xname);
		goto fail;
	}
	if (!(pci_conf_read(pc, pa->pa_tag, PCI_COMMAND_STATUS_REG) &
		PCI_COMMAND_MEM_ENABLE)) {
		printf("%s: failed to enable memory mapping\n",
			sc->sc_dv.dv_xname);
		goto fail;
	}

	if ((error = pci_mapreg_map(pa, WC_PCI_BAR, PCI_MAPREG_TYPE_IO, 0,
		&sc->sc_iot, &sc->sc_ioh, &sc->sc_iob, &sc->sc_iosize)) != 0) {
	    aprint_error("%s: can't map i/o space, error = %d\n",
		sc->sc_dv.dv_xname, error);
	    goto fail;
	}

	sc->sc_pct = pc;
	sc->sc_dmat = pa->pa_dmat;
	sc->sc_pos = sc->sc_dv.dv_unit;
	sc->sc_variety = id->device_info->name;
	sc->sc_flags = id->device_info->flags;
	sc->sc_usecount = 0;

	zaptel_attach_mi(&sc->sc_dv);

	/* Map and establish the interrupt */
	if (pci_intr_map(pa, &ih)) {
	    aprint_error("%s: can't map interrupt\n", sc->sc_dv.dv_xname);
	    goto fail;
	}

	intrstr = pci_intr_string(pc, ih);
	sc->sc_ih = pci_intr_establish(pc, ih, IPL_NET, wcfxo_intr, sc);
	if (sc->sc_ih == NULL) {
	    aprint_error("%s: unable to establish interrupt",
		sc->sc_dv.dv_xname);
	    if (intrstr != NULL)
		aprint_normal(" at %s", intrstr);
	    aprint_normal("\n");
	    goto fail;
	}
	aprint_normal("%s: interrupting at %s\n",sc->sc_dv.dv_xname, intrstr);

	/*
	 * Set up DMA mappings
	 */
	if (!wcfxo_dma_malloc(sc, &sc->sc_read))
	    goto fail;
	if (!wcfxo_dma_malloc(sc, &sc->sc_write))
	    goto fail_1;

	if (wcfxo_initialize(sc)) {
	    aprint_error("%s: unable to initialize\n",
		sc->sc_dv.dv_xname);
	    goto fail_2;
	}

	if (wcfxo_hardware_init(sc)) {
		aprint_error("%s: unable to init hardware\n", 
				sc->sc_dv.dv_xname);
		goto fail_2;
	}

	wcfxo_generate_description(sc);

	/* Initialize buffers to 0 */
	memset((void *) sc->sc_read.chunk, 0, WC_SIZE);
	memset((void *) sc->sc_write.chunk, 0, WC_SIZE);

	wcfxo_enable_interrupts(sc);
	wcfxo_start_dma(sc);

	/* DAA can't be initialized until interrupts are enabled */
	config_interrupts(self, wcfxo_finish_config);

DPRINTF(1,("%s: done\n",__func__));
	return;

 fail_2:
	wcfxo_dma_free(sc, &sc->sc_write);
 fail_1:
	wcfxo_dma_free(sc, &sc->sc_read);
 fail:
	return;
}

static void
wcfxo_finish_config(struct device *self)
{
	struct wcfxo_softc *sc = (struct wcfxo_softc *)self;

	/* Initialize DAA after it's started */
	if (wcfxo_init_daa(sc)) {
		aprint_error("%s: Failed to init DAA, giving up...\n",
			sc->sc_dv.dv_xname);
		return;
	}

	wcfxo_set_daa_mode(sc);
	wcfxo_generate_description(sc);
}

static int
wcfxo_detach(struct device* self, int flags)
{
	struct wcfxo_softc *sc = (struct wcfxo_softc *) self;

	if (sc) {
		/* don't unload if module is in use */
		if (sc->sc_usecount)
			return (EBUSY);
	}

	wcfxo_disable_interrupts(sc);
	if (sc->sc_ih != NULL)
	    pci_intr_disestablish(sc->sc_pct, sc->sc_ih);
	wcfxo_stop_dma(sc);
	wcfxo_reset_tdm(sc);

	/* free read/write dma buffers */
	wcfxo_dma_free(sc, &sc->sc_read);
	wcfxo_dma_free(sc, &sc->sc_write);

	/* Reset PCI chip and registers */
	WC_WRITE1(WC_CNTL, 0x0e);

	wcfxo_release(sc);

	if (sc->sc_iosize)
	    bus_space_unmap(sc->sc_iot, sc->sc_ioh, sc->sc_iosize);

	if (card_configuration) {
	    free(card_configuration, M_DEVBUF);
	    card_configuration = NULL;
	}

	aprint_naive("%s: detach!\n",sc->sc_dv.dv_xname);

	return (0);
}

/* wcfxo_intr: interrupt service routine */
static int
wcfxo_intr(void *arg)
{
	struct wcfxo_softc *sc = arg;
	unsigned char ints = 0xff;
	unsigned char b;
#ifdef DEBUG_RING
	static int oldb = 0;
	static int oldcnt = 0;
#endif

	ints = WC_READ1(WC_INTSTAT);
	DPRINTF(2,("%s: 0x%x\n",__func__, ints));
	WC_WRITE1(WC_INTSTAT, ints);

	if (!ints)
		return 0;

	if (ints & 0x0c) {
		/* if there is an rx interrupt pending, transmitprep
		   looks to see if there is anything to transmit and
		   returns by itself if there is nothing. */
		wcfxo_receiveprep(sc, ints);
		wcfxo_transmitprep(sc, ints);
	}

	if (ints & 0x10) {
		/* Stop DMA, wait for watchdog */
		printf("TDM PCI Master abort\n");
		wcfxo_stop_dma(sc);
		return 1;
	}

	if (ints & 0x20) {
		printf("PCI Target abort\n");
		return 1;
	}

	if (batterycheck) {
	    /* Check for BATTERY from register and debounce for 8 ms */
	    b = sc->sc_readregs[0xc] & 0xf;
	    if (!b) {
		sc->sc_nobatttimer++;
		if (sc->sc_battery && !sc->sc_battdebounce) {
		    DPRINTF(1,("%s: NO BATTERY\n",sc->sc_dv.dv_xname));
		    sc->sc_battery = 0;
#ifdef JAPAN
		    if (!sc->sc_ohdebounce && sc->sc_offhook) {
			zt_hooksig(&sc->sc_chan, ZT_RXSIG_ONHOOK);
			DPRINTF(1,("%s: signalled on-hook\n",
			    sc->sc_dv.dv_xname));
#ifdef ZERO_BATT_RING
			sc->sc_onhook++;
#endif
		    }
else
			zt_hooksig(&sc->sc_chan, ZT_RXSIG_ONHOOK);
#endif /* JAPAN */
		    sc->sc_battdebounce = BATT_DEBOUNCE;
		} else if (!sc->sc_battery)
		    sc->sc_battdebounce = BATT_DEBOUNCE;
		if ((sc->sc_nobatttimer > 5000) &&
#ifdef ZERO_BATT_RING
		    !(sc->sc_readregs[0x05] & 0x04) &&
#endif
		    (!sc->sc_span.alarms)) {
		    sc->sc_span.alarms = ZT_ALARM_RED;
		    zt_alarm_notify(&sc->sc_span);
		    wcfxo_generate_description(sc);
		}
	    } else if (b == 0xf) {
		if (!sc->sc_battery && !sc->sc_battdebounce) {
		    DPRINTF(1,("%s: BATTERY!\n", sc->sc_dv.dv_xname));
#ifdef ZERO_BATT_RING
		    if (sc->sc_onhook) {
			sc->sc_onhook = 0;
			zt_hooksig(&sc->sc_chan, ZT_RXSIG_OFFHOOK);
			DPRINTF(1,("%s: Signalled off-hook\n",
			    sc->sc_dv.dv_xname));
		    }
#else
		    zt_hooksig(&sc->sc_chan, ZT_RXSIG_OFFHOOK);
#endif /* ZERO_BATT_RING */
		    sc->sc_battery = 1;
		    sc->sc_nobatttimer = 0;
		    sc->sc_battdebounce = BATT_DEBOUNCE;
		    if (sc->sc_span.alarms) {
			sc->sc_span.alarms = 0;
			zt_alarm_notify(&sc->sc_span);
			wcfxo_generate_description(sc);
		    }
		} else if (sc->sc_battery)
		    sc->sc_battdebounce = BATT_DEBOUNCE;
	    } else {
		/* It's something else... */
		sc->sc_battdebounce = BATT_DEBOUNCE;
	    }
	    if (sc->sc_battdebounce)
		sc->sc_battdebounce--;
#ifdef JAPAN
	    if (sc->sc_ohdebounce)
		sc->sc_ohdebounce--;
#endif
	} else {
	    /* We don't want to check battery status.  It could be
	     * useful for old pulse telephone lines, to prevent the
	     * driver from making a wrong hangup.  So, just reset all
	     * alarms on the line.
	     */
	    if (sc->sc_span.alarms) {
		sc->sc_span.alarms = 0;
		zt_alarm_notify(&sc->sc_span);
		wcfxo_generate_description(sc);
	    }
	}

	return 1;
}

int wcfxo_open(struct zt_chan *chan)
{
	struct wcfxo_softc *sc = chan->pvt;
	DPRINTF(1,("%s\n",__func__));
	if (sc->sc_dead)
		return ENODEV;
	sc->sc_usecount++;
	return 0;
}

static int wcfxo_watchdog(struct zt_span *span, int event)
{
	printf("WCFXO: Restarting DMA\n");
	wcfxo_restart_dma(span->pvt);
	return 0;
}

static int wcfxo_close(struct zt_chan *chan)
{
	DPRINTF(1,("%s\n",__func__));
	struct wcfxo_softc *sc = chan->pvt;
	sc->sc_usecount--;
	/* Actualy it is useless in BSD version but let it remains */
	/* If we're dead, release us now */
	if (!sc->sc_usecount && sc->sc_dead) 
		wcfxo_release(sc);
	return 0;
}

static void wcfxo_release(struct wcfxo_softc *sc)
{
	DPRINTF(1,("%s\n",__func__));
	zt_unregister(&sc->sc_span);
	printf("%s: freed a wildcard\n", sc->sc_dv.dv_xname);
}

static int wcfxo_hooksig(struct zt_chan *chan, zt_txsig_t txsig)
{
	struct wcfxo_softc *sc = chan->pvt;
	int reg=0;

	DPRINTF(1,("%s\n",__func__));
	switch(txsig) {
	case ZT_TXSIG_START:
	case ZT_TXSIG_OFFHOOK:
		/* Take off hook and enable nornal mode reception.  This
		   must be done in two steps because of a hardware bug. */
		reg = sc->sc_readregs[0x5] & ~0x08;
		wcfxo_setreg(sc, 5, reg);

		reg = reg | 0x1;
		wcfxo_setreg(sc, 0x5, reg);
		sc->sc_offhook = 1;
#ifdef JAPAN
		sc->sc_battery = 1;
		sc->sc_battdebounce = BATT_DEBOUNCE;
		sc->sc_ohdebounce = OH_DEBOUNCE;
#endif
		break;
	case ZT_TXSIG_ONHOOK:
		/* Put on hook and enable on-hook line monitor */
		reg = sc->sc_readregs[0x5] & 0xfe;
		wcfxo_setreg(sc, 0x5, reg);

		reg = reg | 0x08;
		wcfxo_setreg(sc, 0x5, reg);

		sc->sc_offhook = 0;
		/* Don't accept a ring for another 1000ms */
		sc->sc_ringdebounce = 1000;
#ifdef JAPAN
		sc->sc_ohdebounce = OH_DEBOUNCE;
#endif
 		break;
	default:
		printf("%s: Can't set tx state to %d\n", sc->sc_dv.dv_xname,
				txsig);
	}
	DPRINTF(1,("%s: Setting hook state to %d (%02x)\n",
				sc->sc_dv.dv_xname, txsig, reg));
 
	return 0;
}
